from typing import TYPE_CHECKING, cast

from ape.api import UpstreamProvider
from ape_ethereum.provider import Web3Provider
from web3 import HTTPProvider, Web3
from web3.exceptions import ExtraDataLengthError
from web3.gas_strategies.rpc import rpc_gas_price_strategy
from web3.middleware.validation import MAX_EXTRADATA_LENGTH

from .client import DrpcClient

try:
    from web3.middleware import ExtraDataToPOAMiddleware  # type: ignore
except ImportError:
    from web3.middleware import geth_poa_middleware as ExtraDataToPOAMiddleware  # type: ignore

if TYPE_CHECKING:
    from .config import DrpcConfig


class DrpcProvider(Web3Provider, UpstreamProvider, DrpcClient):
    """
    A web3 provider using dRPC caching RPC proxy.

    Docs: https://docs.alchemy.com/alchemy/
    """

    @property
    def config(self) -> "DrpcConfig":
        return cast("DrpcConfig", self.config_manager.get_config("drpc"))

    @property
    def uri(self) -> str:
        if not (
            (networks := self.supported_ecosystems.get(self.network.ecosystem.name))
            and (network := networks.get(self.network.name))
        ):
            raise ValueError(
                f"Unsupported ecosystem/network: {self.network.ecosystem.name}:{self.network.name}"
            )

        uri = f"{str(self.config.host).rstrip('/')}/{network.short_names[0]}"

        if api_key := self.config.api_key:
            return f"{uri}/{api_key}"

        return uri

    @property
    def http_uri(self) -> str:
        # NOTE: Overriding `Web3Provider.http_uri` implementation
        return self.uri

    @property
    def ws_uri(self):
        if self.config.use_tor_proxy:
            raise ValueError("This provider does not support websockets over Tor")

        # NOTE: Overriding `Web3Provider.http_uri` implementation
        return self.uri.replace("https://", "wss://")

    def connect(self):
        self._web3 = Web3(HTTPProvider(self.uri))
        is_poa = None
        try:
            # Any chain that *began* as PoA needs the middleware for pre-merge blocks
            avalanche = 43114
            base = 8453
            optimism = 10
            polygon = 137
            polygon_amoy = 80002

            if self._web3.eth.chain_id in (avalanche, base, optimism, polygon, polygon_amoy):
                self._web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
                is_poa = True

            self._web3.eth.set_gas_price_strategy(rpc_gas_price_strategy)
        except Exception:
            is_poa = None

        if is_poa is None:
            # Check if is PoA but just wasn't as such yet.
            # NOTE: We have to check both earliest and latest
            #   because if the chain was _ever_ PoA, we need
            #   this middleware.
            for option in ("earliest", "latest"):
                try:
                    block = self.web3.eth.get_block(option)  # type: ignore[arg-type]
                except ExtraDataLengthError:
                    is_poa = True
                    break
                else:
                    is_poa = (
                        "proofOfAuthorityData" in block
                        or len(block.get("extraData", "")) > MAX_EXTRADATA_LENGTH
                    )
                    if is_poa:
                        break

            if is_poa and ExtraDataToPOAMiddleware not in self.web3.middleware_onion:
                self.web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

    def disconnect(self):
        self._web3 = None



from functools import cache

import requests
import yaml
from ape.types import HexInt
from ape.utils.basemodel import ManagerAccessMixin
from pydantic import BaseModel, Field

CHAIN_META_URL = "https://raw.githubusercontent.com/drpcorg/public/refs/heads/main/chains.yaml"


class ChainSettings(BaseModel):
    mev_critical: bool = Field(alias="mev-critical", default=False)


class ChainMeta(BaseModel):
    id: str
    code: str
    chain_id: HexInt = Field(alias="chain-id")

    short_names: list[str] = Field(alias="short-names")
    settings: ChainSettings = ChainSettings()

    grpc_id: int = Field(alias="grpcId")


class DrpcClient(ManagerAccessMixin):
    @classmethod
    @cache
    def get_supported_ecosystems(cls) -> dict[str, dict[str, ChainMeta]]:
        if (response := requests.get(CHAIN_META_URL)).status_code >= 400:
            response.raise_for_status()

        return {
            ecosystem.name: {
                network.name: chain
                for chain in map(ChainMeta.model_validate, protocol["chains"])
                # NOTE: Check that dRPC-supported network is in Ape's network list
                if (network := ecosystem.networks.get(chain.id.lower()))
                and network.chain_id == chain.chain_id
            }
            for protocol in yaml.safe_load(response.text)["chain-settings"]["protocols"]
            if protocol["type"] in ("eth",)
            # NOTE: Check that dRPC-supported ecosystem is in Ape's ecosystem list
            and (ecosystem := cls.network_manager.ecosystems.get(protocol["id"]))
        }

    @property
    def supported_ecosystems(self) -> dict[str, dict[str, ChainMeta]]:
        return self.__class__.get_supported_ecosystems()



from datetime import datetime, timedelta, timezone

import pytest

from ape_drpc.provider import DrpcProvider

# NOTE: These cause problems but ostensibly work sometimes
SKIP = [
    "berachain",
    "blast",
    "celo",
    "cronos",
    "filecoin",
    "kroma",
    "moonbeam",
    "sophon",
    "taiko",
    "zksync",
]


@pytest.mark.parametrize(
    "ecosystem", (e for e in DrpcProvider.get_supported_ecosystems() if e not in SKIP)
)
def test_connection(networks, ecosystem):
    for network in DrpcProvider.get_supported_ecosystems()[ecosystem]:
        with networks.parse_network_choice(f"{ecosystem}:{network}:drpc") as provider:
            assert isinstance(provider, DrpcProvider)
            assert abs(
                provider.get_block("latest").datetime - datetime.now(timezone.utc)
            ) < timedelta(minutes=5)



from ape.api import PluginConfig
from pydantic import AnyUrl
from pydantic_settings import SettingsConfigDict


class DrpcConfig(PluginConfig):
    # NOTE: dRPC NodeCloud instance
    host: AnyUrl = AnyUrl("https://lb.drpc.live")
    api_key: str | None = None

    model_config = SettingsConfigDict(env_prefix="APE_DRPC_")



from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from .config import DrpcConfig

    return DrpcConfig


@plugins.register(plugins.ProviderPlugin)
def providers():
    from .provider import DrpcProvider

    supported_ecosystems = DrpcProvider.get_supported_ecosystems()
    for ecosystem_name, supported_networks in supported_ecosystems.items():
        for network_name in supported_networks:
            yield ecosystem_name, network_name, DrpcProvider
